#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <assert.h>
#include <GL/glew.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

void display(void);
void draw_rect(void);
void reshape(int x, int y);
void keydown(unsigned char key, int x, int y);
void glut_printf(float x, float y, const char *fmt, ...);
int next_pow2(int x);

unsigned int fbo, rtex;
int rtex_width, rtex_height;	/* rendertarget texture size */
int width, height;				/* window size */

int main(int argc, char **argv)
{
	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutCreateWindow("FBO resize example");
	width = glutGet(GLUT_WINDOW_WIDTH);
	height = glutGet(GLUT_WINDOW_HEIGHT);

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keydown);

	glewInit();

	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	/* create the FBO and its render target texture */
	rtex_width = next_pow2(width);
	rtex_height = next_pow2(height);

	glGenFramebuffers(1, &fbo);
	glBindFramebuffer(GL_FRAMEBUFFER, fbo);

	glGenTextures(1, &rtex);
	glBindTexture(GL_TEXTURE_2D, rtex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, rtex_width, rtex_height, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, rtex, 0);

	glBindTexture(GL_TEXTURE_2D, 0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	glutMainLoop();
	return 0;
}

void display(void)
{
	/* ----- draw something into the texture ----- */
	glBindFramebuffer(GL_FRAMEBUFFER, fbo);
	glClearColor(0.1, 0.2, 0.4, 1);
	glClear(GL_COLOR_BUFFER_BIT);

	glBegin(GL_QUADS);
	glColor3f(0.8, 0.3, 0.2);
	glVertex2f(-1, -1);
	glVertex2f(1, -1);
	glVertex2f(1, 1);
	glVertex2f(-1, 1);
	glEnd();

	glColor3f(1, 1, 0.2);
	glut_printf(-0.9, 0.1, "window size: %dx%d", width, height);
	glut_printf(-0.9, -0.1, "texture size: %dx%d", rtex_width, rtex_height);


	/* ----- revert to the regular framebuffer ----- */
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glClearColor(0, 0, 0, 1);
	glClear(GL_COLOR_BUFFER_BIT);

	/* now display the texture fullscreen */
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, rtex);

	/* draw fullscreen quad with the *useful* part of the texture */
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glScalef((float)width / (float)rtex_width, (float)height / (float)rtex_height, 1);

	draw_rect();
	glPopMatrix();

	/* also show a small window with the full texture near the top-right */
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(0.7, 0.7, 0);
	glScalef(0.25, 0.25, 1);

	draw_rect();

	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);

	/* draw a black rect around the small window to make it more visible */
	glBegin(GL_LINE_LOOP);
	glColor3f(0, 0, 0);
	glVertex2f(-1, -1);
	glVertex2f(1, -1);
	glVertex2f(1, 1);
	glVertex2f(-1, 1);
	glEnd();
	glPopMatrix();

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void draw_rect(void)
{
	glBegin(GL_QUADS);
	glColor3f(1, 1, 1);
	glTexCoord2f(0, 0); glVertex2f(-1, -1);
	glTexCoord2f(1, 0); glVertex2f(1, -1);
	glTexCoord2f(1, 1); glVertex2f(1, 1);
	glTexCoord2f(0, 1); glVertex2f(-1, 1);
	glEnd();
}

void reshape(int x, int y)
{
	int new_rtex_width, new_rtex_height;

	glViewport(0, 0, x, y);

	width = x;
	height = y;
	new_rtex_width = next_pow2(width);
	new_rtex_height = next_pow2(height);

	if(rtex_width != new_rtex_width || rtex_height != new_rtex_height) {
		/* resize the texture */
		rtex_width = new_rtex_width;
		rtex_height = new_rtex_height;

		glBindTexture(GL_TEXTURE_2D, rtex);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, rtex_width, rtex_height, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
		glutPostRedisplay();
	}
}

void keydown(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);
	}
}

void glut_printf(float x, float y, const char *fmt, ...)
{
	float line_width = sqrt(width * width + height * height) / 400.0 + 0.5;
	va_list ap;
	static char buf[4096];
	char *ptr = buf;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof buf - 1, fmt, ap);
	va_end(ap);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glScalef(1.0, (float)width / (float)height, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(x, y, 0);
	glScalef(0.0008, 0.0008, 1);

	glLineWidth(line_width);
	while(*ptr) {
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *ptr++);
	}

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

int next_pow2(int x)
{
	--x;
	x |= x >> 1;
	x |= x >> 2;
	x |= x >> 4;
	x |= x >> 8;
	x |= x >> 16;
	return x + 1;
}
